==> sparrow-parse-1.0.5.tar.gz: sparrow-parse-1.0.5/sparrow_parse/vllm/mlx_inference.py <==
from mlx_vlm import load, generate
from mlx_vlm.prompt_utils import apply_chat_template
from mlx_vlm.utils import load_image
from sparrow_parse.vllm.inference_base import ModelInference
import os
import json
from rich import print


class MLXInference(ModelInference):
    """
        A class for performing inference using the MLX model.
        Handles image preprocessing, response formatting, and model interaction.
        """

    def __init__(self, model_name):
        """
        Initialize the inference class with the given model name.

        :param model_name: Name of the model to load.
        """
        self.model_name = model_name
        print(f"MLXInference initialized for model: {model_name}")


    @staticmethod
    def _load_model_and_processor(model_name):
        """
        Load the model and processor for inference.

        :param model_name: Name of the model to load.
        :return: Tuple containing the loaded model and processor.
        """
        model, processor = load(model_name)
        print(f"Loaded model: {model_name}")
        return model, processor


    def process_response(self, output_text):
        """
        Process and clean the model's raw output to format as JSON.
        """
        try:
            # Check if we have markdown code block markers
            if "```" in output_text:
                # Handle markdown-formatted output
                json_start = output_text.find("```json")
                if json_start != -1:
                    # Extract content between ```json and ```
                    content = output_text[json_start + 7:]
                    json_end = content.rfind("```")
                    if json_end != -1:
                        content = content[:json_end].strip()
                        formatted_json = json.loads(content)
                        return json.dumps(formatted_json, indent=2)

            # Handle raw JSON (no markdown formatting)
            # First try to find JSON array or object patterns
            for pattern in [r'\[\s*\{.*\}\s*\]', r'\{.*\}']:
                import re
                matches = re.search(pattern, output_text, re.DOTALL)
                if matches:
                    potential_json = matches.group(0)
                    try:
                        formatted_json = json.loads(potential_json)
                        return json.dumps(formatted_json, indent=2)
                    except:
                        pass

            # Last resort: try to parse the whole text as JSON
            formatted_json = json.loads(output_text.strip())
            return json.dumps(formatted_json, indent=2)

        except Exception as e:
            print(f"Failed to parse JSON: {e}")
            return output_text


    def load_image_data(self, image_filepath, max_width=1250, max_height=1750):
        """
        Load and resize image while maintaining its aspect ratio.

        :param image_filepath: Path to the image file.
        :param max_width: Maximum allowed width of the image.
        :param max_height: Maximum allowed height of the image.
        :return: Tuple containing the image object and its new dimensions.
        """
        image = load_image(image_filepath)  # Assuming load_image is defined elsewhere
        width, height = image.size

        # Calculate new dimensions while maintaining the aspect ratio
        if width > max_width or height > max_height:
            aspect_ratio = width / height
            new_width = min(max_width, int(max_height * aspect_ratio))
            new_height = min(max_height, int(max_width / aspect_ratio))
            return image, new_width, new_height

        return image, width, height


    def inference(self, input_data, mode=None):
        """
        Perform inference on input data using the specified model.

        :param input_data: A list of dictionaries containing image file paths and text inputs.
        :param mode: Optional mode for inference ("static" for simple JSON output).
        :return: List of processed model responses.
        """
        # Handle static mode
        if mode == "static":
            return [self.get_simple_json()]

        # Load the model and processor
        model, processor = self._load_model_and_processor(self.model_name)
        config = model.config
        
        # Determine if we're doing text-only or image-based inference
        is_text_only = input_data[0].get("file_path") is None
        
        if is_text_only:
            # Text-only inference
            messages = input_data[0]["text_input"]
            response = self._generate_text_response(model, processor, config, messages)
            results = [response]
        else:
            # Image-based inference
            file_paths = self._extract_file_paths(input_data)
            results = self._process_images(model, processor, config, file_paths, input_data)
        
        return results

    def _generate_text_response(self, model, processor, config, messages):
        """
        Generate a text response for text-only inputs.
        
        :param model: The loaded model
        :param processor: The loaded processor
        :param config: Model configuration
        :param messages: Input messages
        :return: Generated response
        """
        prompt = apply_chat_template(processor, config, messages)
        response =  generate(
            model,
            processor,
            prompt,
            max_tokens=4000,
            temperature=0.0,
            verbose=False
        )
        print("Inference completed successfully")
        return response

    def _process_images(self, model, processor, config, file_paths, input_data):
        """
        Process images and generate responses for each.
        
        :param model: The loaded model
        :param processor: The loaded processor
        :param config: Model configuration
        :param file_paths: List of image file paths
        :param input_data: Original input data
        :return: List of processed responses
        """
        results = []
        for file_path in file_paths:
            image, width, height = self.load_image_data(file_path)
            
            # Prepare messages based on model type
            messages = self._prepare_messages(input_data, file_path)
            
            # Generate and process response
            prompt = apply_chat_template(processor, config, messages)
            response = generate(
                model,
                processor,
                prompt,
                image,
                resize_shape=(width, height),
                max_tokens=4000,
                temperature=0.0,
                verbose=False
            )
            results.append(self.process_response(response))
            print(f"Inference completed successfully for: {file_path}")
        
        return results

    def _prepare_messages(self, input_data, file_path):
        """
        Prepare the appropriate messages based on the model type.
        
        :param input_data: Original input data
        :param file_path: Current file path being processed
        :return: Properly formatted messages
        """
        if "mistral" in self.model_name.lower():
            return input_data[0]["text_input"]
        else:
            return [
                {"role": "system", "content": "You are an expert at extracting structured text from image documents."},
                {"role": "user", "content": input_data[0]["text_input"]},
            ]

    @staticmethod
    def _extract_file_paths(input_data):
        """
        Extract and resolve absolute file paths from input data.

        :param input_data: List of dictionaries containing image file paths.
        :return: List of absolute file paths.
        """
        return [
            os.path.abspath(file_path)
            for data in input_data
            for file_path in data.get("file_path", [])
        ]

==> sparrow-parse-1.0.6.tar.gz: sparrow-parse-1.0.6/sparrow_parse/vllm/mlx_inference.py <==
from mlx_vlm import load, generate
from mlx_vlm.prompt_utils import apply_chat_template
from mlx_vlm.utils import load_image
from sparrow_parse.vllm.inference_base import ModelInference
import os
import json, re
from rich import print


class MLXInference(ModelInference):
    """
        A class for performing inference using the MLX model.
        Handles image preprocessing, response formatting, and model interaction.
        """

    def __init__(self, model_name):
        """
        Initialize the inference class with the given model name.

        :param model_name: Name of the model to load.
        """
        self.model_name = model_name
        print(f"MLXInference initialized for model: {model_name}")


    @staticmethod
    def _load_model_and_processor(model_name):
        """
        Load the model and processor for inference.

        :param model_name: Name of the model to load.
        :return: Tuple containing the loaded model and processor.
        """
        model, processor = load(model_name)
        print(f"Loaded model: {model_name}")
        return model, processor


    def process_response(self, output_text):
        """
        Process and clean the model's raw output to format as JSON.
        """
        try:
            # Check if we have markdown code block markers
            if "```" in output_text:
                # Handle markdown-formatted output
                json_start = output_text.find("```json")
                if json_start != -1:
                    # Extract content between ```json and ```
                    content = output_text[json_start + 7:]
                    json_end = content.rfind("```")
                    if json_end != -1:
                        content = content[:json_end].strip()
                        formatted_json = json.loads(content)
                        return json.dumps(formatted_json, indent=2)

            # Handle raw JSON (no markdown formatting)
            # First try to find JSON array or object patterns
            for pattern in [r'\[\s*\{.*\}\s*\]', r'\{.*\}']:
                import re
                matches = re.search(pattern, output_text, re.DOTALL)
                if matches:
                    potential_json = matches.group(0)
                    try:
                        formatted_json = json.loads(potential_json)
                        return json.dumps(formatted_json, indent=2)
                    except:
                        pass

            # Last resort: try to parse the whole text as JSON
            formatted_json = json.loads(output_text.strip())
            return json.dumps(formatted_json, indent=2)

        except Exception as e:
            print(f"Failed to parse JSON: {e}")
            return output_text


    def load_image_data(self, image_filepath, max_width=1250, max_height=1750):
        """
        Load and resize image while maintaining its aspect ratio.

        :param image_filepath: Path to the image file.
        :param max_width: Maximum allowed width of the image.
        :param max_height: Maximum allowed height of the image.
        :return: Tuple containing the image object and its new dimensions.
        """
        image = load_image(image_filepath)  # Assuming load_image is defined elsewhere
        width, height = image.size

        # Calculate new dimensions while maintaining the aspect ratio
        if width > max_width or height > max_height:
            aspect_ratio = width / height
            new_width = min(max_width, int(max_height * aspect_ratio))
            new_height = min(max_height, int(max_width / aspect_ratio))
            return image, new_width, new_height

        return image, width, height


    def inference(self, input_data, apply_annotation=False, mode=None):
        """
        Perform inference on input data using the specified model.

        :param input_data: A list of dictionaries containing image file paths and text inputs.
        :param apply_annotation: Optional flag to apply annotations to the output.
        :param mode: Optional mode for inference ("static" for simple JSON output).
        :return: List of processed model responses.
        """
        # Handle static mode
        if mode == "static":
            return [self.get_simple_json()]

        # Load the model and processor
        model, processor = self._load_model_and_processor(self.model_name)
        config = model.config
        
        # Determine if we're doing text-only or image-based inference
        is_text_only = input_data[0].get("file_path") is None
        
        if is_text_only:
            # Text-only inference
            messages = input_data[0]["text_input"]
            response = self._generate_text_response(model, processor, config, messages)
            results = [response]
        else:
            # Image-based inference
            file_paths = self._extract_file_paths(input_data)
            results = self._process_images(model, processor, config, file_paths, input_data, apply_annotation)
        
        return results

    def _generate_text_response(self, model, processor, config, messages):
        """
        Generate a text response for text-only inputs.
        
        :param model: The loaded model
        :param processor: The loaded processor
        :param config: Model configuration
        :param messages: Input messages
        :return: Generated response
        """
        prompt = apply_chat_template(processor, config, messages)
        response =  generate(
            model,
            processor,
            prompt,
            max_tokens=4000,
            temperature=0.0,
            verbose=False
        )
        print("Inference completed successfully")
        return response

    def _process_images(self, model, processor, config, file_paths, input_data, apply_annotation):
        """
        Process images and generate responses for each.
        
        :param model: The loaded model
        :param processor: The loaded processor
        :param config: Model configuration
        :param file_paths: List of image file paths
        :param input_data: Original input data
        :param apply_annotation: Flag to apply annotations
        :return: List of processed responses
        """
        results = []
        for file_path in file_paths:
            image, width, height = self.load_image_data(file_path)
            
            # Prepare messages based on model type
            messages = self._prepare_messages(input_data, apply_annotation)
            
            # Generate and process response
            prompt = apply_chat_template(processor, config, messages)
            response, _ = generate(
                model,
                processor,
                prompt,
                image,
                resize_shape=(width, height),
                max_tokens=4000,
                temperature=0.0,
                verbose=False
            )
            results.append(self.process_response(response))
            print(f"Inference completed successfully for: {file_path}")
        
        return results


    def transform_query_with_bbox(self, text_input):
        """
        Transform JSON schema in text_input to include value, bbox, and confidence.
        Works with both array and object JSON structures.

        Args:
            text_input (str): The input text containing a JSON schema

        Returns:
            str: Text with transformed JSON including value, bbox, and confidence
        """
        # Split text into parts - find the JSON portion between "retrieve" and "return response"
        retrieve_pattern = r'retrieve\s+'
        return_pattern = r'\.\s+return\s+response'

        retrieve_match = re.search(retrieve_pattern, text_input)
        return_match = re.search(return_pattern, text_input)

        if not retrieve_match or not return_match:
            return text_input  # Return original if pattern not found

        json_start = retrieve_match.end()
        json_end = return_match.start()

        prefix = text_input[:json_start]
        json_str = text_input[json_start:json_end].strip()
        suffix = text_input[json_end:]

        # Parse and transform the JSON
        try:
            # Handle single quotes if needed
            json_str = json_str.replace("'", '"')

            json_obj = json.loads(json_str)
            transformed_json = self.transform_query_structure(json_obj)
            transformed_json_str = json.dumps(transformed_json)

            # Rebuild the text
            result = prefix + transformed_json_str + suffix

            return result
        except json.JSONDecodeError as e:
            print(f"Error parsing JSON: {e}")
            return text_input  # Return original if parsing fails


    def transform_query_structure(self, json_obj):
        """
        Transform each field in the JSON structure to include value, bbox, and confidence.
        Handles both array and object formats recursively.
        """
        if isinstance(json_obj, list):
            # Handle array format
            return [self.transform_query_structure(item) for item in json_obj]
        elif isinstance(json_obj, dict):
            # Handle object format
            result = {}
            for key, value in json_obj.items():
                if isinstance(value, (dict, list)):
                    # Recursively transform nested objects or arrays
                    result[key] = self.transform_query_structure(value)
                else:
                    # Transform simple value to object with value, bbox, and confidence
                    result[key] = {
                        "value": value,
                        "bbox": ["float", "float", "float", "float"],
                        "confidence": "float"
                    }
            return result
        else:
            # For primitive values, no transformation needed
            return json_obj


    def _prepare_messages(self, input_data, apply_annotation):
        """
        Prepare the appropriate messages based on the model type.
        
        :param input_data: Original input data
        :param apply_annotation: Flag to apply annotations
        :return: Properly formatted messages
        """
        if "mistral" in self.model_name.lower():
            return input_data[0]["text_input"]
        elif "qwen" in self.model_name.lower():
            if apply_annotation:
                system_prompt = {"role": "system", "content": "You are an expert at extracting text from images. "
                                                              "For each item in the table, provide separate bounding boxes for each field. "
                                                              "All coordinates should be in pixels relative to the original image. Format your response in JSON."}
                user_prompt = {"role": "user", "content": self.transform_query_with_bbox(input_data[0]["text_input"])}
                return [system_prompt, user_prompt]
            return [
                {"role": "system", "content": "You are an expert at extracting structured text from image documents."},
                {"role": "user", "content": input_data[0]["text_input"]},
            ]
        else:
            raise ValueError("Unsupported model type. Please use either Mistral or Qwen.")

    @staticmethod
    def _extract_file_paths(input_data):
        """
        Extract and resolve absolute file paths from input data.

        :param input_data: List of dictionaries containing image file paths.
        :return: List of absolute file paths.
        """
        return [
            os.path.abspath(file_path)
            for data in input_data
            for file_path in data.get("file_path", [])
        ]
